#!/usr/bin/env python
import os
import copy
import json
import unittest
import manifest

from emodpy.utils import EradicationBambooBuilds
from emodpy.bamboo import get_model_files
import emod_api.campaign as camp

current_directory = os.path.dirname(os.path.realpath(__file__))


class TestIntervention(unittest.TestCase):
    """
    Base Test class for TBHIV intervention
    """
    @classmethod
    def setUpClass(cls) -> None:
        cls.output = os.path.join(current_directory, 'output')
        if not os.path.isdir(cls.output):
            os.mkdir(cls.output)

        cls.schema = manifest.schema_file
        if not os.path.isfile(cls.schema):
            plan = EradicationBambooBuilds.TBHIV
            print("Retrieving Eradication and schema.json from Bamboo...")
            get_model_files(plan, manifest)

    def setUp(self) -> None:
        print(f"\n{self._testMethodName} started...")
        camp.pubsub_signals_pubbing = []
        camp.pubsub_signals_subbing = []

    def trigger_treatment_list_test(self, intervention_fn, intervention_class_name,
                                    campaign_filename="camp.json"):
        camp.schema_path = self.schema
        expected_trigger_treatment_list = ["NewInfection"]
        trigger_treatment_list = copy.deepcopy(expected_trigger_treatment_list)
        event = intervention_fn(camp, trigger_treatment_list)
        camp.add(event, first=True)
        camp_filename = os.path.join(self.output, campaign_filename)
        manifest.delete_existing_file(camp_filename)
        camp.save(camp_filename)
        self.assertTrue(os.path.isfile(camp_filename))

        with open(camp_filename, 'r') as file:
            camp_file = json.load(file)
        self.assertEqual(camp_file["Events"][0]["Event_Coordinator_Config"]["Intervention_Config"]
                         ["Trigger_Condition_List"], expected_trigger_treatment_list)
        self.assertEqual(camp.pubsub_signals_subbing,
                         copy.deepcopy(expected_trigger_treatment_list) + ['Blackout'])
        self.assertEqual(camp_file["Events"][0]["Event_Coordinator_Config"]["Intervention_Config"]
                         ["Actual_IndividualIntervention_Config"]["class"], intervention_class_name)

    def trigger_treatment_list_with_drug_name_test(self, intervention_fn, drug_name,
                                                   intervention_class_name, campaign_filename="camp.json"):
        camp.schema_path = self.schema
        expected_trigger_treatment_list = ["NewInfection"]
        trigger_treatment_list = copy.deepcopy(expected_trigger_treatment_list)
        event = intervention_fn(camp, trigger_treatment_list, drug_name)
        camp.add(event, first=True)
        camp_filename = os.path.join(self.output, campaign_filename)
        manifest.delete_existing_file(camp_filename)
        camp.save(camp_filename)
        self.assertTrue(os.path.isfile(camp_filename))

        with open(camp_filename, 'r') as file:
            camp_file = json.load(file)
        self.assertEqual(camp_file["Events"][0]["Event_Coordinator_Config"]["Intervention_Config"]
                         ["Trigger_Condition_List"], expected_trigger_treatment_list)
        self.assertEqual(camp.pubsub_signals_subbing,
                         copy.deepcopy(expected_trigger_treatment_list) + ['Blackout'])
        self.assertEqual(camp_file["Events"][0]["Event_Coordinator_Config"]["Intervention_Config"]
                         ["Actual_IndividualIntervention_Config"]["class"], intervention_class_name)
        self.assertEqual(camp_file["Events"][0]["Event_Coordinator_Config"]["Intervention_Config"]
                         ["Actual_IndividualIntervention_Config"]["TB_Drug_Name"],
                         drug_name)

    def trigger_treatment_list_with_output_event_test(self, intervention_fn, output_event,
                                                      intervention_class_name, campaign_filename="camp.json"):
        camp.schema_path = self.schema
        expected_trigger_treatment_list = ["NewInfection"]
        trigger_treatment_list = copy.deepcopy(expected_trigger_treatment_list)
        event = intervention_fn(camp, trigger_treatment_list, output_event)
        camp.add(event, first=True)
        camp_filename = os.path.join(self.output, campaign_filename)
        manifest.delete_existing_file(camp_filename)
        camp.save(camp_filename)
        self.assertTrue(os.path.isfile(camp_filename))

        with open(camp_filename, 'r') as file:
            camp_file = json.load(file)
        self.assertEqual(camp_file["Events"][0]["Event_Coordinator_Config"]["Intervention_Config"]
                         ["Trigger_Condition_List"], expected_trigger_treatment_list)
        self.assertEqual(camp.pubsub_signals_subbing,
                         copy.deepcopy(expected_trigger_treatment_list) + ['Blackout'])
        self.assertEqual(camp_file["Events"][0]["Event_Coordinator_Config"]["Intervention_Config"]
                         ["Actual_IndividualIntervention_Config"]["class"], intervention_class_name)
        self.assertEqual(camp_file["Events"][0]["Event_Coordinator_Config"]["Intervention_Config"]
                         ["Actual_IndividualIntervention_Config"]["Actual_IndividualIntervention_Event"],
                         output_event)

    def trigger_treatment_list_with_properties_test(self, intervention_fn, property_to_change,
                                                    final_prop_value, intervention_class_name,
                                                    campaign_filename="camp.json"):
        camp.schema_path = self.schema
        expected_trigger_treatment_list = ["NewInfection"]
        trigger_treatment_list = copy.deepcopy(expected_trigger_treatment_list)
        event = intervention_fn(camp, trigger_treatment_list, property_to_change, final_prop_value)
        camp.add(event, first=True)
        camp_filename = os.path.join(self.output, campaign_filename)
        manifest.delete_existing_file(camp_filename)
        camp.save(camp_filename)
        self.assertTrue(os.path.isfile(camp_filename))

        with open(camp_filename, 'r') as file:
            camp_file = json.load(file)
        self.assertEqual(camp_file["Events"][0]["Event_Coordinator_Config"]["Intervention_Config"]
                         ["Trigger_Condition_List"], expected_trigger_treatment_list)
        self.assertEqual(camp.pubsub_signals_subbing,
                         copy.deepcopy(expected_trigger_treatment_list) + ['Blackout'])
        self.assertEqual(camp_file["Events"][0]["Event_Coordinator_Config"]["Intervention_Config"]
                         ["Actual_IndividualIntervention_Config"]["class"], intervention_class_name)
        self.assertEqual(camp_file["Events"][0]["Event_Coordinator_Config"]["Intervention_Config"]
                         ["Actual_IndividualIntervention_Config"]["Target_Property_Key"],
                         property_to_change)
        self.assertEqual(camp_file["Events"][0]["Event_Coordinator_Config"]["Intervention_Config"]
                         ["Actual_IndividualIntervention_Config"]["Target_Property_Value"],
                         final_prop_value)

    def sensitivity_and_specificity_test(self, intervention_fn, base_sensitivity, base_specificity,
                                         campaign_filename):
        camp.schema_path = self.schema
        expected_trigger_treatment_list = ["NewInfection"]
        trigger_treatment_list = copy.deepcopy(expected_trigger_treatment_list)
        event = intervention_fn(camp, trigger_treatment_list, base_sensitivity=base_sensitivity,
                                base_specificity=base_specificity)
        camp.add(event, first=True)
        camp_filename = os.path.join(self.output, campaign_filename)
        manifest.delete_existing_file(camp_filename)
        camp.save(camp_filename)
        self.assertTrue(os.path.isfile(camp_filename))

        with open(camp_filename, 'r') as file:
            camp_file = json.load(file)
        self.assertEqual(camp_file["Events"][0]["Event_Coordinator_Config"]["Intervention_Config"]
                         ["Actual_IndividualIntervention_Config"]["Base_Sensitivity"], base_sensitivity)
        self.assertEqual(camp_file["Events"][0]["Event_Coordinator_Config"]["Intervention_Config"]
                         ["Actual_IndividualIntervention_Config"]["Base_Specificity"], base_specificity)


